# Copyright (c) OpenMMLab. All rights reserved.
from abc import abstractmethod
from itertools import repeat
from typing import Callable, Iterable, Optional, Tuple

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.nn.modules.conv import _ConvNd

from mmrazor.models.mutables.base_mutable import BaseMutable
from .dynamic_mixins import DynamicChannelMixin


def _ntuple(n: int) -> Callable:  # pragma: no cover
    """Repeat a number n times."""

    def parse(x):
        if isinstance(x, Iterable):
            return tuple(x)
        return tuple(repeat(x, n))

    return parse


def _get_current_kernel_pos(source_kernel_size: int,
                            target_kernel_size: int) -> Tuple[int, int]:
    """Get position of current kernel size.

    Returns:
        Tuple[int, int]: (upper left position, bottom right position)
    """
    assert source_kernel_size >= target_kernel_size, \
        '`source_kernel_size` must greater or equal than `target_kernel_size`'

    center = source_kernel_size >> 1
    current_offset = target_kernel_size >> 1

    start_offset = center - current_offset
    end_offset = center + current_offset + 1

    return start_offset, end_offset


def _get_same_padding(kernel_size: int, n_dims: int) -> Tuple[int]:
    """Get same padding according to kernel size."""
    assert kernel_size & 1
    _pair = _ntuple(n_dims)

    return _pair(kernel_size >> 1)


class DynamicConvMixin(DynamicChannelMixin):
    """A mixin class for Pytorch conv, which can mutate ``in_channels`` and
    ``out_channels``.

    Note:
        All subclass should implement ``conv_func``API.
    """

    @property
    @abstractmethod
    def conv_func(self: _ConvNd):
        """The function that will be used in ``forward_mixin``."""
        pass

    def register_mutable_attr(self, attr, mutable):

        if attr == 'in_channels':
            self._register_mutable_in_channels(mutable)
        elif attr == 'out_channels':
            self._register_mutable_out_channels(mutable)
        else:
            raise NotImplementedError

    def _register_mutable_in_channels(
            self: _ConvNd, mutable_in_channels: BaseMutable) -> None:
        """Mutate ``in_channels`` with given mutable.

        Args:
            mutable_in_channels (BaseMutable): Mutable for controlling
                ``in_channels``.

        Raises:
            ValueError: Error if size of mask if not same as ``in_channels``.
        """
        assert hasattr(self, 'mutable_attrs')
        self.check_mutable_channels(mutable_in_channels)
        mask_size = mutable_in_channels.current_mask.size(0)
        if mask_size != self.in_channels:
            raise ValueError(
                f'Expect mask size of mutable to be {self.in_channels} as '
                f'`in_channels`, but got: {mask_size}.')

        self.mutable_attrs['in_channels'] = mutable_in_channels

    def _register_mutable_out_channels(
            self: _ConvNd, mutable_out_channels: BaseMutable) -> None:
        """Mutate ``out_channels`` with given mutable.

        Args:
            mutable_out_channels (BaseMutable): Mutable for controlling
                ``out_channels``.

        Raises:
            ValueError: Error if size of mask if not same as ``out_channels``.
        """
        assert hasattr(self, 'mutable_attrs')
        self.check_mutable_channels(mutable_out_channels)
        mask_size = mutable_out_channels.current_mask.size(0)
        if mask_size != self.out_channels:
            raise ValueError(
                f'Expect mask size of mutable to be {self.out_channels} as '
                f'`out_channels`, but got: {mask_size}.')

        self.mutable_attrs['out_channels'] = mutable_out_channels

    @property
    def mutable_in_channels(self: _ConvNd) -> Optional[BaseMutable]:
        """Mutable related to input."""
        assert hasattr(self, 'mutable_attrs')
        return getattr(self.mutable_attrs, 'in_channels', None)  # type:ignore

    @property
    def mutable_out_channels(self: _ConvNd) -> Optional[BaseMutable]:
        """Mutable related to output."""
        assert hasattr(self, 'mutable_attrs')
        return getattr(self.mutable_attrs, 'out_channels', None)  # type:ignore

    def get_dynamic_params(
            self: _ConvNd) -> Tuple[Tensor, Optional[Tensor], Tuple[int]]:
        """Get dynamic parameters that will be used in forward process.

        Returns:
            Tuple[Tensor, Optional[Tensor], Tuple[int]]: Sliced weight, bias
                and padding.
        """
        # slice in/out channel of weight according to
        # mutable in_channels/out_channels
        weight, bias = self._get_dynamic_params_by_mutable_channels(
            self.weight, self.bias)
        return weight, bias, self.padding

    def _get_dynamic_params_by_mutable_channels(
            self: _ConvNd, weight: Tensor,
            bias: Optional[Tensor]) -> Tuple[Tensor, Optional[Tensor]]:
        """Get sliced weight and bias according to ``mutable_in_channels`` and
        ``mutable_out_channels``.

        Returns:
            Tuple[Tensor, Optional[Tensor]]: Sliced weight and bias.
        """
        if 'in_channels' not in self.mutable_attrs and \
                'out_channels' not in self.mutable_attrs:
            return weight, bias

        if 'in_channels' in self.mutable_attrs:
            mutable_in_channels = self.mutable_attrs['in_channels']
            in_mask = mutable_in_channels.current_mask.to(weight.device)
        else:
            in_mask = torch.ones(weight.size(1)).bool().to(weight.device)

        if 'out_channels' in self.mutable_attrs:
            mutable_out_channels = self.mutable_attrs['out_channels']
            out_mask = mutable_out_channels.current_mask.to(weight.device)
        else:
            out_mask = torch.ones(weight.size(0)).bool().to(weight.device)

        if self.groups == 1:
            weight = weight[out_mask][:, in_mask]
        elif self.groups == self.in_channels == self.out_channels:
            # depth-wise conv
            weight = weight[out_mask]
        else:
            raise NotImplementedError(
                'Current `ChannelMutator` only support pruning the depth-wise '
                '`nn.Conv2d` or `nn.Conv2d` module whose group number equals '
                f'to one, but got {self.groups}.')
        bias = self.bias[out_mask] if self.bias is not None else None
        return weight, bias

    def forward_mixin(self: _ConvNd, x: Tensor) -> Tensor:
        """Forward of dynamic conv2d OP."""
        groups = self.groups
        if self.groups == self.in_channels == self.out_channels:
            groups = x.size(1)
        weight, bias, padding = self.get_dynamic_params()

        return self.conv_func(x, weight, bias, self.stride, padding,
                              self.dilation, groups)

    def to_static_op(self: _ConvNd) -> nn.Conv2d:
        """Convert dynamic conv2d to :obj:`torch.nn.Conv2d`.

        Returns:
            torch.nn.Conv2d: :obj:`torch.nn.Conv2d` with sliced parameters.
        """
        self.check_if_mutables_fixed()

        weight, bias, padding = self.get_dynamic_params()
        groups = self.groups
        if groups == self.in_channels == self.out_channels and \
                self.mutable_in_channels is not None:
            mutable_in_channels = self.mutable_attrs['in_channels']
            groups = mutable_in_channels.current_mask.sum().item()
        out_channels = weight.size(0)
        in_channels = weight.size(1) * groups

        kernel_size = tuple(weight.shape[2:])

        static_conv = self.static_op_factory(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=self.stride,
            padding=padding,
            padding_mode=self.padding_mode,
            dilation=self.dilation,
            groups=groups,
            bias=True if bias is not None else False)

        static_conv.weight = nn.Parameter(weight)
        if bias is not None:
            static_conv.bias = nn.Parameter(bias)

        return static_conv


class BigNasConvMixin(DynamicConvMixin):
    """A mixin class for Pytorch conv, which can mutate ``in_channels``,
    ``out_channels`` and ``kernel_size``."""

    def register_mutable_attr(self, attr, mutable):

        if attr == 'in_channels':
            self._register_mutable_in_channels(mutable)
        elif attr == 'out_channels':
            self._register_mutable_out_channels(mutable)
        elif attr == 'kernel_size':
            self._register_mutable_kernel_size(mutable)
        else:
            raise NotImplementedError

    def _register_mutable_kernel_size(
            self: _ConvNd, mutable_kernel_size: BaseMutable) -> None:
        """Mutate ``kernel_size`` with given mutable.

        Args:
            mutable_kernel_size (BaseMutable): Mutable for controlling
                ``kernel_size``.

        Note:
            ``kernel_size_seq`` must be provided if ``mutable_kernel_size``
            does not have ``choices`` attribute.

        Raises:
            ValueError: Error if max choice of ``kernel_size_list``
                not same as ``kernel_size``.
        """

        kernel_size_seq = getattr(mutable_kernel_size, 'choices', None)
        if kernel_size_seq is None or len(kernel_size_seq) == 0:
            raise ValueError('kernel size sequence must be provided')
        kernel_size_list = list(sorted(kernel_size_seq))

        _pair = _ntuple(len(self.weight.shape) - 2)
        max_kernel_size = _pair(kernel_size_list[-1])
        if max_kernel_size != self.kernel_size:
            raise ValueError(
                f'Expect max kernel size to be: {self.kernel_size}, '
                f'but got: {max_kernel_size}')

        self.kernel_size_list = kernel_size_list
        self.mutable_attrs['kernel_size'] = mutable_kernel_size

    def get_dynamic_params(
            self: _ConvNd) -> Tuple[Tensor, Optional[Tensor], Tuple[int]]:
        """Get dynamic parameters that will be used in forward process.

        Returns:
            Tuple[Tensor, Optional[Tensor], Tuple[int]]: Sliced weight, bias
                and padding.
        """
        # 1. slice kernel size of weight according to kernel size mutable
        weight, padding = self._get_dynamic_params_by_mutable_kernel_size(
            self.weight)

        # 2. slice in/out channel of weight according to mutable in_channels
        # and mutable out channels.
        weight, bias = self._get_dynamic_params_by_mutable_channels(
            weight, self.bias)
        return weight, bias, padding

    def _get_dynamic_params_by_mutable_kernel_size(
            self: _ConvNd, weight: Tensor) -> Tuple[Tensor, Tuple[int]]:
        """Get sliced weight and bias according to ``mutable_in_channels`` and
        ``mutable_out_channels``."""
        if 'kernel_size' not in self.mutable_attrs \
                or self.kernel_size_list is None:
            return weight, self.padding

        mutable_kernel_size = self.mutable_attrs['kernel_size']
        current_kernel_size = self.get_current_choice(mutable_kernel_size)

        n_dims = len(self.weight.shape) - 2
        current_padding = _get_same_padding(current_kernel_size, n_dims)

        _pair = _ntuple(len(self.weight.shape) - 2)
        if _pair(current_kernel_size) == self.kernel_size:
            return weight, current_padding

        start_offset, end_offset = _get_current_kernel_pos(
            source_kernel_size=self.kernel_size[0],
            target_kernel_size=current_kernel_size)
        current_weight = \
            weight[:, :, start_offset:end_offset, start_offset:end_offset]

        return current_weight, current_padding

    def forward_mixin(self: _ConvNd, x: Tensor) -> Tensor:
        """Forward of dynamic conv2d OP."""
        groups = self.groups
        if self.groups == self.in_channels == self.out_channels:
            groups = x.size(1)
        weight, bias, padding = self.get_dynamic_params()

        return self.conv_func(x, weight, bias, self.stride, padding,
                              self.dilation, groups)


class OFAConvMixin(BigNasConvMixin):
    """A mixin class for Pytorch conv, which can mutate ``in_channels``,
    ``out_channels`` and ``kernel_size``."""

    def _register_mutable_kernel_size(
            self: _ConvNd, mutable_kernel_size: BaseMutable) -> None:
        """Mutate ``kernel_size`` with given mutable and register
        transformation matrix."""
        super()._register_mutable_kernel_size(mutable_kernel_size)
        self._register_trans_matrix()

    def _register_trans_matrix(self: _ConvNd) -> None:
        """Register transformation matrix that used in progressive
        shrinking."""
        assert self.kernel_size_list is not None

        trans_matrix_names = []
        for i in range(len(self.kernel_size_list) - 1, 0, -1):
            source_kernel_size = self.kernel_size_list[i]
            target_kernel_size = self.kernel_size_list[i - 1]
            trans_matrix_name = self._get_trans_matrix_name(
                src=source_kernel_size, tar=target_kernel_size)
            trans_matrix_names.append(trans_matrix_name)
            # TODO support conv1d & conv3d
            trans_matrix = nn.Parameter(torch.eye(target_kernel_size**2))
            self.register_parameter(name=trans_matrix_name, param=trans_matrix)
        self._trans_matrix_names = trans_matrix_names

    @staticmethod
    def _get_trans_matrix_name(src: int, tar: int) -> str:
        """Get name of trans matrix."""
        return f'trans_matrix_{src}to{tar}'

    def _get_dynamic_params_by_mutable_kernel_size(
            self: _ConvNd, weight: Tensor) -> Tuple[Tensor, Tuple[int]]:
        """Get sliced weight and bias according to ``mutable_in_channels`` and
        ``mutable_out_channels``."""

        if 'kernel_size' not in self.mutable_attrs:
            return weight, self.padding

        mutable_kernel_size = self.mutable_attrs['kernel_size']
        current_kernel_size = self.get_current_choice(mutable_kernel_size)

        n_dims = len(self.weight.shape) - 2
        current_padding = _get_same_padding(current_kernel_size, n_dims)

        _pair = _ntuple(len(self.weight.shape) - 2)
        if _pair(current_kernel_size) == self.kernel_size:
            return weight, current_padding

        current_weight = weight[:, :, :, :]
        for i in range(len(self.kernel_size_list) - 1, 0, -1):
            source_kernel_size = self.kernel_size_list[i]
            if source_kernel_size <= current_kernel_size:
                break
            target_kernel_size = self.kernel_size_list[i - 1]
            trans_matrix = getattr(
                self,
                self._get_trans_matrix_name(
                    src=source_kernel_size, tar=target_kernel_size))

            start_offset, end_offset = _get_current_kernel_pos(
                source_kernel_size=source_kernel_size,
                target_kernel_size=target_kernel_size)
            target_weight = current_weight[:, :, start_offset:end_offset,
                                           start_offset:end_offset]
            target_weight = target_weight.reshape(-1, target_kernel_size**2)
            target_weight = F.linear(target_weight, trans_matrix)
            target_weight = target_weight.reshape(
                weight.size(0), weight.size(1), target_kernel_size,
                target_kernel_size)

            current_weight = target_weight

        return current_weight, current_padding



_base_ = [
    'mmseg::_base_/datasets/cityscapes.py',
    'mmseg::_base_/schedules/schedule_80k.py',
    'mmseg::_base_/default_runtime.py'
]

teacher_ckpt = 'https://download.openmmlab.com/mmsegmentation/v0.5/pspnet/pspnet_r101-d8_512x1024_80k_cityscapes/pspnet_r101-d8_512x1024_80k_cityscapes_20200606_112211-e1e1100f.pth'  # noqa: E501
teacher_cfg_path = 'mmseg::pspnet/pspnet_r101-d8_4xb2-80k_cityscapes-512x1024.py'  # noqa: E501
student_cfg_path = 'mmseg::pspnet/pspnet_r18-d8_4xb2-80k_cityscapes-512x1024.py'  # noqa: E501
model = dict(
    _scope_='mmrazor',
    type='SingleTeacherDistill',
    architecture=dict(cfg_path=student_cfg_path, pretrained=False),
    teacher=dict(cfg_path=teacher_cfg_path, pretrained=False),
    teacher_ckpt=teacher_ckpt,
    distiller=dict(
        type='ConfigurableDistiller',
        distill_losses=dict(
            loss_cwd=dict(type='ChannelWiseDivergence', tau=1, loss_weight=5)),
        student_recorders=dict(
            logits=dict(type='ModuleOutputs', source='decode_head.conv_seg')),
        teacher_recorders=dict(
            logits=dict(type='ModuleOutputs', source='decode_head.conv_seg')),
        loss_forward_mappings=dict(
            loss_cwd=dict(
                preds_S=dict(from_student=True, recorder='logits'),
                preds_T=dict(from_student=False, recorder='logits')))))

find_unused_parameters = True

val_cfg = dict(_delete_=True, type='mmrazor.SingleTeacherDistillValLoop')



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

import torch
import torch.nn as nn

from mmrazor.registry import MODELS
from .base_connector import BaseConnector


@MODELS.register_module()
class Paraphraser(BaseConnector):
    """Paraphrasing Complex Network: Network Compression via Factor Transfer,
    NeurIPS 2018. https://arxiv.org/pdf/1802.04977.pdf.

    teacher connector of FT.

    Args:
        in_channel ([int]): number of input channels.
        out_channel ([int]): number of output channels.
        use_bn (bool, optional): Defaults to False.
        phase (str, optional): Training phase. Defaults to 'pretrain'.
        use_bn (Optional[bool], optional): use BN or not. Defaults to False.
        init_cfg (Optional[Dict], optional): The weight initialized config for
            :class:`BaseModule`. Defaults to None.
    """

    def __init__(self,
                 in_channel: int,
                 out_channel: int,
                 phase='pretrain',
                 use_bn: Optional[bool] = False,
                 init_cfg: Optional[Dict] = None) -> None:

        super(Paraphraser, self).__init__(init_cfg)
        self._build_modules(in_channel, out_channel, use_bn)

        assert phase in ['pretrain', 'train'], f'Unexpect `phase`: {phase}'
        self.phase = phase

    def _build_modules(self,
                       in_channel: int,
                       out_channel: int,
                       use_bn: Optional[bool] = False) -> None:
        """A helper func to build internal modules.

        Args:
            in_channel (int): input channels
            out_channel (int): output channels
            use_bn (Optional[bool], optional): use BN or not.
                Defaults to False.
        """

        self.encoder = nn.Sequential(
            nn.Conv2d(in_channel, in_channel, 3, 1, 1),
            nn.BatchNorm2d(in_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True),
            nn.Conv2d(in_channel, out_channel, 3, 1, 1),
            nn.BatchNorm2d(out_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True),
            nn.Conv2d(out_channel, out_channel, 3, 1, 1),
            nn.BatchNorm2d(out_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True))
        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(out_channel, out_channel, 3, 1, 1),
            nn.BatchNorm2d(out_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True),
            nn.ConvTranspose2d(out_channel, in_channel, 3, 1, 1),
            nn.BatchNorm2d(in_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True),
            nn.ConvTranspose2d(in_channel, in_channel, 3, 1, 1),
            nn.BatchNorm2d(in_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True))

    def forward_train(self, x: torch.Tensor) -> torch.Tensor:
        """Forward func for training."""
        with torch.no_grad():
            factor = self.encoder(x)
        return factor

    def forward_pretrain(self, t_feat: torch.Tensor) -> torch.Tensor:
        """Forward func for pretraining."""
        factor = self.encoder(t_feat)
        t_feat_rec = self.decoder(factor)

        return t_feat_rec

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Omitted."""
        if self.phase == 'train':
            return self.forward_train(x)
        elif self.phase == 'pretrain':
            return self.forward_pretrain(x)
        else:
            raise NotImplementedError(
                f'phase: `{self.phase}` is not supported.')


@MODELS.register_module()
class Translator(BaseConnector):
    """Paraphrasing Complex Network: Network Compression via Factor Transfer,
    NeurIPS 2018. https://arxiv.org/pdf/1802.04977.pdf.

    student connector of FT.

    Args:
        in_channel ([int]): number of input channels.
        out_channel ([int]): number of output channels.
        use_bn (bool, optional): Defaults to False.
        init_cfg (Optional[Dict], optional): The weight initialized config for
            :class:`BaseModule`. Defaults to None.
    """

    def __init__(self,
                 in_channel: int,
                 out_channel: int,
                 use_bn: Optional[bool] = True,
                 init_cfg: Optional[Dict] = None) -> None:
        super(Translator, self).__init__(init_cfg)
        self.encoder = nn.Sequential(
            nn.Conv2d(in_channel, in_channel, 3, 1, 1),
            nn.BatchNorm2d(in_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True),
            nn.Conv2d(in_channel, out_channel, 3, 1, 1),
            nn.BatchNorm2d(out_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True),
            nn.Conv2d(out_channel, out_channel, 3, 1, 1),
            nn.BatchNorm2d(out_channel) if use_bn else nn.Sequential(),
            nn.LeakyReLU(0.1, inplace=True))

    def forward_train(self, x: torch.Tensor) -> torch.Tensor:
        """Forward func for training."""
        return self.encoder(x)



# Copyright (c) OpenMMLab. All rights reserved.
from mmengine.logging import MMLogger
from mmengine.optim import OptimWrapper
from torch.nn import Module


def reinitialize_optim_wrapper_count_status(model: Module,
                                            optim_wrapper: OptimWrapper,
                                            accumulative_counts: int,
                                            verbose: bool = True) -> None:
    if verbose:
        logger = MMLogger.get_current_instance()
        logger.warning('Reinitialize count status of optim wrapper')

    original_max_iters = \
        optim_wrapper.message_hub.runtime_info['max_iters']
    new_max_iters = original_max_iters * accumulative_counts
    original_init_iters = \
        optim_wrapper.message_hub.runtime_info['iter']
    new_init_iters = original_init_iters * accumulative_counts

    if verbose:
        logger.info(f'original `init_iters`: {original_init_iters}, '
                    f'new `init_iters`: {new_init_iters}; '
                    f'orginal `max_iters`: {original_max_iters}, '
                    f'new `max_iters`: {new_max_iters}')

    optim_wrapper.initialize_count_status(
        model=model, init_counts=new_init_iters, max_counts=new_max_iters)



# dataset settings
dataset_type = 'mmcls.CIFAR10'
preprocess_cfg = dict(
    # RGB format normalization parameters
    mean=[125.307, 122.961, 113.8575],
    std=[51.5865, 50.847, 51.255],
    # loaded images are already RGB format
    to_rgb=False)

train_pipeline = [
    dict(type='mmcls.RandomCrop', crop_size=32, padding=4),
    dict(type='mmcls.RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='mmcls.PackClsInputs'),
]

test_pipeline = [
    dict(type='mmcls.PackClsInputs'),
]

train_dataloader = dict(
    batch_size=64,
    num_workers=4,
    dataset=dict(
        type=dataset_type,
        data_prefix='data/cifar10',
        indices=-25000,
        test_mode=False,
        pipeline=train_pipeline),
    sampler=dict(type='mmcls.DefaultSampler', shuffle=True),
    persistent_workers=True,
)

val_dataloader = dict(
    batch_size=128,
    num_workers=4,
    dataset=dict(
        type=dataset_type,
        data_prefix='data/cifar10/',
        test_mode=True,
        pipeline=test_pipeline),
    sampler=dict(type='mmcls.DefaultSampler', shuffle=False),
    persistent_workers=True,
)
val_evaluator = dict(type='mmcls.Accuracy', topk=(1, ))

test_dataloader = val_dataloader
test_evaluator = val_evaluator

# optimizer
optim_wrapper = dict(
    architecture=dict(
        type='mmcls.SGD', lr=0.025, momentum=0.9, weight_decay=3e-4),
    mutator=dict(type='mmcls.Adam', lr=3e-4, weight_decay=1e-3),
    clip_grad=dict(max_norm=5, norm_type=2))

# leanring policy
# TODO support different optim use different scheduler (wait mmengine)
param_scheduler = [
    dict(
        type='mmcls.CosineAnnealingLR',
        T_max=50,
        eta_min=1e-3,
        begin=0,
        end=50),
]
# param_scheduler = dict(
#     architecture = dict(
#         type='mmcls.CosineAnnealingLR',
#         T_max=50,
#         eta_min=1e-3,
#         begin=0,
#         end=50),
#     mutator = dict(
#         type='mmcls.ConstantLR',
#         factor=1,
#         begin=0,
#         end=50))

# train, val, test setting
# TODO split cifar dataset
train_cfg = dict(
    type='mmrazor.DartsEpochBasedTrainLoop',
    mutator_dataloader=dict(
        batch_size=64,
        num_workers=4,
        dataset=dict(
            type=dataset_type,
            data_prefix='data/cifar10',
            indices=25000,
            test_mode=False,
            pipeline=train_pipeline),
        sampler=dict(type='mmcls.DefaultSampler', shuffle=True),
        persistent_workers=True,
    ),
    max_epochs=50)

val_cfg = dict()  # validate each epoch
test_cfg = dict()  # dataset settings



_base_ = [
    'mmdet::_base_/datasets/coco_detection.py',
    'mmdet::_base_/schedules/schedule_1x.py',
    'mmdet::_base_/default_runtime.py'
]

# default_scope = 'mmrazor'
teacher_ckpt = 'faster_rcnn_r101_fpn_2x_coco_bbox_mAP-0.398_20200504_210455-1d2dac9c.pth'  # noqa: E501
model = dict(
    _scope_='mmrazor',
    type='FpnTeacherDistill',
    architecture=dict(
        cfg_path='mmdet::faster_rcnn/faster_rcnn_r50_fpn_1x_coco.py',
        pretrained=False),
    teacher=dict(
        cfg_path='mmdet::faster_rcnn/faster_rcnn_r101_fpn_2x_coco.py',
        pretrained=False),
    teacher_ckpt=teacher_ckpt,
    distiller=dict(
        type='ConfigurableDistiller',
        student_recorders=dict(fpn=dict(type='ModuleOutputs', source='neck')),
        teacher_recorders=dict(fpn=dict(type='ModuleOutputs', source='neck')),
        distill_losses=dict(
            loss_cwd_fpn0=dict(
                type='ChannelWiseDivergence', tau=1, loss_weight=10),
            loss_cwd_fpn1=dict(
                type='ChannelWiseDivergence', tau=1, loss_weight=10),
            loss_cwd_fpn2=dict(
                type='ChannelWiseDivergence', tau=1, loss_weight=10),
            loss_cwd_fpn3=dict(
                type='ChannelWiseDivergence', tau=1, loss_weight=10),
            loss_cwd_fpn4=dict(
                type='ChannelWiseDivergence', tau=1, loss_weight=10)),
        loss_forward_mappings=dict(
            loss_cwd_fpn0=dict(
                preds_S=dict(from_student=True, recorder='fpn', data_idx=0),
                preds_T=dict(from_student=False, recorder='fpn', data_idx=0)),
            loss_cwd_fpn1=dict(
                preds_S=dict(from_student=True, recorder='fpn', data_idx=1),
                preds_T=dict(from_student=False, recorder='fpn', data_idx=1)),
            loss_cwd_fpn2=dict(
                preds_S=dict(from_student=True, recorder='fpn', data_idx=2),
                preds_T=dict(from_student=False, recorder='fpn', data_idx=2)),
            loss_cwd_fpn3=dict(
                preds_S=dict(from_student=True, recorder='fpn', data_idx=3),
                preds_T=dict(from_student=False, recorder='fpn', data_idx=3)),
            loss_cwd_fpn4=dict(
                preds_S=dict(from_student=True, recorder='fpn', data_idx=4),
                preds_T=dict(from_student=False, recorder='fpn',
                             data_idx=4)))))

find_unused_parameters = True

val_cfg = dict(_delete_=True, type='mmrazor.SingleTeacherDistillValLoop')



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Tuple

import torch
import torch.nn as nn
from mmcv.cnn import NonLocal2d

from mmrazor.registry import MODELS
from .base_connector import BaseConnector


class NonLocal2dMaxpoolNstride(NonLocal2d):
    """Nonlocal block for 2-dimension inputs, with a configurable
    maxpool_stride.

    This module is proposed in
    "Non-local Neural Networks"
    Paper reference: https://arxiv.org/abs/1711.07971
    Code reference: https://github.com/AlexHex7/Non-local_pytorch

    Args:
        in_channels (int): Channels of the input feature map.
        reduction (int): Channel reduction ratio. Defaults to 2.
        conv_cfg (dict): The config dict for convolution layers.
            Defaults to `nn.Conv2d`.
        norm_cfg (dict): The config dict for normalization layers.
            Defaults to `BN`. (This parameter is only applicable to conv_out.)
        mode (str): Options are `gaussian`, `concatenation`,
            `embedded_gaussian` and `dot_product`. Default: dot_product.
        sub_sample (bool): Whether to apply max pooling after pairwise
            function (Note that the `sub_sample` is applied on spatial only).
            Default: False.
        maxpool_stride (int): The stride of the maxpooling module.
            Defaults to 2.
        zeros_init (bool): Whether to use zero to initialize weights of
            `conv_out`. Defaults to True.
    """

    def __init__(self,
                 in_channels: int,
                 reduction: int = 2,
                 conv_cfg: Dict = dict(type='Conv2d'),
                 norm_cfg: Dict = dict(type='BN'),
                 mode: str = 'embedded_gaussian',
                 sub_sample: bool = False,
                 maxpool_stride: int = 2,
                 zeros_init: bool = True,
                 **kwargs) -> None:
        """Inits the NonLocal2dMaxpoolNstride module."""
        super().__init__(
            in_channels=in_channels,
            sub_sample=sub_sample,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            reduction=reduction,
            mode=mode,
            zeros_init=zeros_init,
            **kwargs)
        self.norm_cfg = norm_cfg

        if sub_sample:
            max_pool_layer = nn.MaxPool2d(
                kernel_size=(maxpool_stride, maxpool_stride))
            self.g: nn.Sequential = nn.Sequential(self.g, max_pool_layer)
            if self.mode != 'gaussian':
                self.phi: nn.Sequential = nn.Sequential(
                    self.phi, max_pool_layer)
            else:
                self.phi = max_pool_layer


@MODELS.register_module()
class FBKDStudentConnector(BaseConnector):
    """Improve Object Detection with Feature-based Knowledge Distillation:
    Towards Accurate and Efficient Detectors, ICLR2021.
    https://openreview.net/pdf?id=uKhGRvM8QNH.

    Student connector for FBKD.

    Args:
        in_channels (int): Channels of the input feature map.
        reduction (int): Channel reduction ratio. Defaults to 2.
        conv_cfg (dict): The config dict for convolution layers.
            Defaults to `nn.Conv2d`.
        norm_cfg (dict): The config dict for normalization layers.
            Defaults to `BN`. (This parameter is only applicable to conv_out.)
        mode (str): Options are `gaussian`, `concatenation`,
            `embedded_gaussian` and `dot_product`. Default: dot_product.
        sub_sample (bool): Whether to apply max pooling after pairwise
            function (Note that the `sub_sample` is applied on spatial only).
            Default: False.
        maxpool_stride (int): The stride of the maxpooling module.
            Defaults to 2.
        zeros_init (bool): Whether to use zero to initialize weights of
            `conv_out`. Defaults to True.
        spatial_T (float): Temperature used in spatial-wise pooling.
            Defaults to 0.5.
        channel_T (float): Temperature used in channel-wise pooling.
            Defaults to 0.5.
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(self,
                 in_channels: int,
                 reduction: int = 2,
                 conv_cfg: Dict = dict(type='Conv2d'),
                 norm_cfg: Dict = dict(type='BN'),
                 mode: str = 'dot_product',
                 sub_sample: bool = False,
                 maxpool_stride: int = 2,
                 zeros_init: bool = True,
                 spatial_T: float = 0.5,
                 channel_T: float = 0.5,
                 init_cfg: Optional[Dict] = None,
                 **kwargs) -> None:
        """Inits the FBKDStuConnector."""
        super().__init__(init_cfg)
        self.channel_wise_adaptation = nn.Linear(in_channels, in_channels)

        self.spatial_wise_adaptation = nn.Conv2d(
            1, 1, kernel_size=3, stride=1, padding=1)

        self.adaptation_layers = nn.Conv2d(
            in_channels, in_channels, kernel_size=1, stride=1, padding=0)

        self.student_non_local = NonLocal2dMaxpoolNstride(
            in_channels=in_channels,
            reduction=reduction,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            mode=mode,
            sub_sample=sub_sample,
            maxpool_stride=maxpool_stride,
            zeros_init=zeros_init,
            **kwargs)

        self.non_local_adaptation = nn.Conv2d(
            in_channels, in_channels, kernel_size=1, stride=1, padding=0)

        self.in_channels = in_channels
        self.spatial_T = spatial_T
        self.channel_T = channel_T

    def forward_train(self, x: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """Frorward function for training.

        Args:
            x (torch.Tensor): Input student features.

        Returns:
            s_spatial_mask (torch.Tensor): Student spatial-wise mask.
            s_channel_mask (torch.Tensor): Student channel-wise mask.
            s_feat_adapt (torch.Tensor): Adaptative student feature.
            s_channel_pool_adapt (torch.Tensor): Student feature which through
                channel-wise pooling and adaptation_layers.
            s_spatial_pool_adapt (torch.Tensor): Student feature which through
                spatial-wise pooling and adaptation_layers.
            s_relation_adapt (torch.Tensor): Adaptative student relations.
        """
        # Calculate spatial-wise mask.
        s_spatial_mask = torch.mean(torch.abs(x), [1], keepdim=True)
        size = s_spatial_mask.size()
        s_spatial_mask = s_spatial_mask.view(x.size(0), -1)

        # Soften or sharpen the spatial-wise mask by temperature.
        s_spatial_mask = torch.softmax(
            s_spatial_mask / self.spatial_T, dim=1) * size[-1] * size[-2]
        s_spatial_mask = s_spatial_mask.view(size)

        # Calculate channel-wise mask.
        s_channel_mask = torch.mean(torch.abs(x), [2, 3], keepdim=True)
        channel_mask_size = s_channel_mask.size()
        s_channel_mask = s_channel_mask.view(x.size(0), -1)

        # Soften or sharpen the channel-wise mask by temperature.
        s_channel_mask = torch.softmax(
            s_channel_mask / self.channel_T, dim=1) * self.in_channels
        s_channel_mask = s_channel_mask.view(channel_mask_size)

        # Adaptative and pool student feature through channel-wise.
        s_feat_adapt = self.adaptation_layers(x)
        s_channel_pool_adapt = self.channel_wise_adaptation(
            torch.mean(x, [2, 3]))

        # Adaptative and pool student feature through spatial-wise.
        s_spatial_pool = torch.mean(x, [1]).view(
            x.size(0), 1, x.size(2), x.size(3))
        s_spatial_pool_adapt = self.spatial_wise_adaptation(s_spatial_pool)

        # Calculate non_local_adaptation.
        s_relation = self.student_non_local(x)
        s_relation_adapt = self.non_local_adaptation(s_relation)

        return (s_spatial_mask, s_channel_mask, s_channel_pool_adapt,
                s_spatial_pool_adapt, s_relation_adapt, s_feat_adapt)


@MODELS.register_module()
class FBKDTeacherConnector(BaseConnector):
    """Improve Object Detection with Feature-based Knowledge Distillation:
    Towards Accurate and Efficient Detectors, ICLR2021.
    https://openreview.net/pdf?id=uKhGRvM8QNH.

    Teacher connector for FBKD.

    Args:
        in_channels (int): Channels of the input feature map.
        reduction (int): Channel reduction ratio. Defaults to 2.
        conv_cfg (dict): The config dict for convolution layers.
            Defaults to `nn.Conv2d`.
        norm_cfg (dict): The config dict for normalization layers.
            Defaults to `BN`. (This parameter is only applicable to conv_out.)
        mode (str): Options are `gaussian`, `concatenation`,
            `embedded_gaussian` and `dot_product`. Default: dot_product.
        sub_sample (bool): Whether to apply max pooling after pairwise
            function (Note that the `sub_sample` is applied on spatial only).
            Default: False.
        maxpool_stride (int): The stride of the maxpooling module.
            Defaults to 2.
        zeros_init (bool): Whether to use zero to initialize weights of
            `conv_out`. Defaults to True.
        spatial_T (float): Temperature used in spatial-wise pooling.
            Defaults to 0.5.
        channel_T (float): Temperature used in channel-wise pooling.
            Defaults to 0.5.
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(self,
                 in_channels,
                 reduction=2,
                 conv_cfg: Dict = dict(type='Conv2d'),
                 norm_cfg: Dict = dict(type='BN'),
                 mode: str = 'dot_product',
                 sub_sample: bool = False,
                 maxpool_stride: int = 2,
                 zeros_init: bool = True,
                 spatial_T: float = 0.5,
                 channel_T: float = 0.5,
                 init_cfg: Optional[Dict] = None,
                 **kwargs) -> None:
        super().__init__(init_cfg)
        self.teacher_non_local = NonLocal2dMaxpoolNstride(
            in_channels=in_channels,
            reduction=reduction,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            mode=mode,
            sub_sample=sub_sample,
            maxpool_stride=maxpool_stride,
            zeros_init=zeros_init,
            **kwargs)

        self.in_channels = in_channels
        self.spatial_T = spatial_T
        self.channel_T = channel_T

    def forward_train(self, x: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """Frorward function for training.

        Args:
            x (torch.Tensor): Input teacher features.

        Returns:
            t_spatial_mask (torch.Tensor): Teacher spatial-wise mask.
            t_channel_mask (torch.Tensor): Teacher channel-wise mask.
            t_spatial_pool (torch.Tensor): Teacher features which through
                spatial-wise pooling.
            t_relation (torch.Tensor): Teacher relation matrix.
        """
        # Calculate spatial-wise mask.
        t_spatial_mask = torch.mean(torch.abs(x), [1], keepdim=True)
        size = t_spatial_mask.size()
        t_spatial_mask = t_spatial_mask.view(x.size(0), -1)

        # Soften or sharpen the spatial-wise mask by temperature.
        t_spatial_mask = torch.softmax(
            t_spatial_mask / self.spatial_T, dim=1) * size[-1] * size[-2]
        t_spatial_mask = t_spatial_mask.view(size)

        # Calculate channel-wise mask.
        t_channel_mask = torch.mean(torch.abs(x), [2, 3], keepdim=True)
        channel_mask_size = t_channel_mask.size()
        t_channel_mask = t_channel_mask.view(x.size(0), -1)

        # Soften or sharpen the channel-wise mask by temperature.
        t_channel_mask = torch.softmax(
            t_channel_mask / self.channel_T, dim=1) * self.in_channels
        t_channel_mask = t_channel_mask.view(channel_mask_size)

        # Adaptative and pool student feature through spatial-wise.
        t_spatial_pool = torch.mean(x, [1]).view(
            x.size(0), 1, x.size(2), x.size(3))

        # Calculate non_local relation.
        t_relation = self.teacher_non_local(x)

        return (t_spatial_mask, t_channel_mask, t_spatial_pool, t_relation, x)



# Copyright (c) OpenMMLab. All rights reserved.
from .base_graph import BaseGraph, BaseNode
from .module_graph import ModuleGraph, ModuleNode

__all__ = ['BaseGraph', 'BaseNode', 'ModuleNode', 'ModuleGraph']



_base_ = [
    'mmcls::_base_/datasets/cifar10_bs16.py',
    'mmcls::_base_/schedules/cifar10_bs128.py',
    'mmcls::_base_/default_runtime.py'
]

model = dict(
    _scope_='mmrazor',
    type='OverhaulFeatureDistillation',
    data_preprocessor=dict(
        type='ImgDataPreprocessor',
        # RGB format normalization parameters
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        # convert image from BGR to RGB
        bgr_to_rgb=True),
    architecture=dict(
        cfg_path=  # noqa: E251
        'mmrazor::vanilla/mmcls/wide-resnet/wrn16-w2_b16x8_cifar10.py',
        pretrained=False),
    teacher=dict(
        cfg_path=  # noqa: E251
        'mmrazor::vanilla/mmcls/wide-resnet/wrn28-w4_b16x8_cifar10.py',
        pretrained=False),
    teacher_ckpt=  # noqa: E251
    'https://download.openmmlab.com/mmrazor/v1/wide_resnet/wrn28_4_b16x8_cifar10_20220831_173536-d6f8725c.pth',  # noqa: E501
    calculate_student_loss=True,
    student_trainable=True,
    distiller=dict(
        type='OFDDistiller',
        student_recorders=dict(
            bb_1=dict(type='ModuleOutputs', source='backbone.layer2.0.bn1'),
            bb_2=dict(type='ModuleOutputs', source='backbone.layer3.0.bn1'),
            bb_3=dict(type='ModuleOutputs', source='backbone.bn1')),
        teacher_recorders=dict(
            bb_1=dict(type='ModuleOutputs', source='backbone.layer2.0.bn1'),
            bb_2=dict(type='ModuleOutputs', source='backbone.layer3.0.bn1'),
            bb_3=dict(type='ModuleOutputs', source='backbone.bn1')),
        distill_losses=dict(
            loss_1=dict(type='OFDLoss', loss_weight=0.25),
            loss_2=dict(type='OFDLoss', loss_weight=0.5),
            loss_3=dict(type='OFDLoss', loss_weight=1.0)),
        connectors=dict(
            loss_1_sfeat=dict(
                type='ConvModuleConncetor',
                in_channel=32,
                out_channel=64,
                norm_cfg=dict(type='BN'),
                act_cfg=None),
            loss_1_tfeat=dict(type='OFDTeacherConnector'),
            loss_2_sfeat=dict(
                type='ConvModuleConncetor',
                in_channel=64,
                out_channel=128,
                norm_cfg=dict(type='BN'),
                act_cfg=None),
            loss_2_tfeat=dict(type='OFDTeacherConnector'),
            loss_3_sfeat=dict(
                type='ConvModuleConncetor',
                in_channel=128,
                out_channel=256,
                norm_cfg=dict(type='BN'),
                act_cfg=None),
            loss_3_tfeat=dict(type='OFDTeacherConnector')),
        loss_forward_mappings=dict(
            loss_1=dict(
                s_feature=dict(
                    from_student=True,
                    recorder='bb_1',
                    connector='loss_1_sfeat'),
                t_feature=dict(
                    from_student=False,
                    recorder='bb_1',
                    connector='loss_1_tfeat'),
            ),
            loss_2=dict(
                s_feature=dict(
                    from_student=True,
                    recorder='bb_2',
                    connector='loss_2_sfeat'),
                t_feature=dict(
                    from_student=False,
                    recorder='bb_2',
                    connector='loss_2_tfeat'),
            ),
            loss_3=dict(
                s_feature=dict(
                    from_student=True,
                    recorder='bb_3',
                    connector='loss_3_sfeat'),
                t_feature=dict(
                    from_student=False,
                    recorder='bb_3',
                    connector='loss_3_tfeat'),
            ),
        )))

find_unused_parameters = True

val_cfg = dict(_delete_=True, type='mmrazor.SingleTeacherDistillValLoop')



# Copyright (c) OpenMMLab. All rights reserved.
from abc import abstractmethod
from typing import Any, Dict, List, Optional

from ..base_mutable import CHOICE_TYPE, CHOSEN_TYPE, BaseMutable


class MutableModule(BaseMutable[CHOICE_TYPE, CHOSEN_TYPE]):
    """Base Class for mutables. Mutable means a searchable module widely used
    in Neural Architecture Search(NAS).

    It mainly consists of some optional operations, and achieving
    searchable function by handling choice with ``MUTATOR``.

    All subclass should implement the following APIs:

    - ``forward()``
    - ``fix_chosen()``
    - ``choices()``

    Args:
        module_kwargs (dict[str, dict], optional): Module initialization named
            arguments. Defaults to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(self,
                 module_kwargs: Optional[Dict[str, Dict]] = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)

        self.module_kwargs = module_kwargs

    @property
    @abstractmethod
    def choices(self) -> List[CHOICE_TYPE]:
        """list: all choices.  All subclasses must implement this method."""

    @abstractmethod
    def forward(self, x: Any) -> Any:
        """Forward computation."""

    @property
    def num_choices(self) -> int:
        """Number of choices."""
        return len(self.choices)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Dict

from mmrazor.registry import MODELS
from ...mutables import OneShotMutableModule
from .module_mutator import ModuleMutator


@MODELS.register_module()
class OneShotModuleMutator(ModuleMutator):
    """One-shot mutable based mutator.

    Examples:
        >>> mutator = OneShotModuleMutator()
        >>> mutator.mutable_class_type
        <class 'mmrazor.models.mutables.oneshot_mutable.OneShotMutable'>

        >>> # Assume that a toy model consists of three mutabels
        >>> # whose name are op1,op2,op3.
        >>> # Each mutable contains 4 choices: choice1, choice2,
        >>> # choice3 and choice4.
        >>> supernet = ToyModel()
        >>> name2module = dict(supernet.named_modules())
        >>> [name for name, module in name2module.items() if isinstance(module, mutator.mutable_class_type)]  # noqa E501
        ['op1', 'op2', 'op3']

        >>> mutator.prepare_from_supernet(supernet)
        >>> mutator.search_groups
        {0: [op1], 1: [op2], 2: [op3]}

        >>> random_choices = mutator.sample_choices()
        {0: 'choice1', 1: 'choice2', 2: 'choice3'}
        >>> mutator.set_subnet(random_choices)

        >>> supernet.op1.current_choice
        'choice1'
        >>> supernet.op2.current_choice
        'choice2'
        >>> supernet.op3.current_choice
        'choice3'

        >>> random_choices_ = mutator.sample_choices()
        {0: 'choice3', 1: 'choice2', 2: 'choice1'}
        >>> mutator.set_subnet(random_choices_)

        >>> supernet.op1.current_choice
        'choice3'
        >>> supernet.op2.current_choice
        'choice2'
        >>> supernet.op3.current_choice
        'choice1'
    """

    def sample_choices(self) -> Dict[int, Any]:
        """Sampling by search groups.

        The sampling result of the first mutable of each group is the sampling
        result of this group.

        Returns:
            Dict[int, Any]: Random choices dict.
        """
        random_choices = dict()
        for group_id, modules in self.search_groups.items():
            random_choices[group_id] = modules[0].sample_choice()

        return random_choices

    def set_choices(self, choices: Dict[int, Any]) -> None:
        """Set mutables' current choice according to choices sample by
        :func:`sample_choices`.

        Args:
            choices (Dict[int, Any]): Choices dict. The key is group_id in
                search groups, and the value is the sampling results
                corresponding to this group.
        """
        for group_id, modules in self.search_groups.items():
            choice = choices[group_id]
            for module in modules:
                module.current_choice = choice

    @property
    def mutable_class_type(self):
        """One-shot mutable class type.

        Returns:
            Type[OneShotMutableModule]: Class type of one-shot mutable.
        """
        return OneShotMutableModule



_base_ = [
    'mmdet::_base_/models/faster-rcnn_r50_fpn.py',
    'mmdet::_base_/datasets/coco_detection.py',
    'mmdet::_base_/schedules/schedule_1x.py',
    'mmdet::_base_/default_runtime.py',
    'mmrazor::_base_/nas_backbones/spos_shufflenet_supernet.py'
]

norm_cfg = dict(type='SyncBN', requires_grad=True)

supernet = _base_.model

supernet.backbone = _base_.nas_backbone
supernet.backbone.norm_cfg = norm_cfg
supernet.backbone.out_indices = (0, 1, 2, 3)
supernet.backbone.with_last_layer = False

supernet.neck.norm_cfg = norm_cfg
supernet.neck.in_channels = [64, 160, 320, 640]

supernet.roi_head.bbox_head.norm_cfg = norm_cfg
supernet.roi_head.bbox_head.type = 'Shared4Conv1FCBBoxHead'

model = dict(
    _delete_=True,
    type='mmrazor.SPOS',
    architecture=supernet,
    mutator=dict(type='mmrazor.OneShotModuleMutator'))

find_unused_parameters = True



# Copyright (c) OpenMMLab. All rights reserved.
from .registry import (DATA_SAMPLERS, DATASETS, HOOKS, LOOPS, METRICS,
                       MODEL_WRAPPERS, MODELS, OPTIM_WRAPPER_CONSTRUCTORS,
                       OPTIM_WRAPPERS, OPTIMIZERS, PARAM_SCHEDULERS,
                       RUNNER_CONSTRUCTORS, RUNNERS, TASK_UTILS, TRANSFORMS,
                       VISBACKENDS, VISUALIZERS, WEIGHT_INITIALIZERS)

__all__ = [
    'RUNNERS', 'RUNNER_CONSTRUCTORS', 'HOOKS', 'DATASETS', 'DATA_SAMPLERS',
    'TRANSFORMS', 'MODELS', 'WEIGHT_INITIALIZERS', 'OPTIMIZERS',
    'OPTIM_WRAPPERS', 'OPTIM_WRAPPER_CONSTRUCTORS', 'TASK_UTILS',
    'PARAM_SCHEDULERS', 'METRICS', 'MODEL_WRAPPERS', 'LOOPS', 'VISBACKENDS',
    'VISUALIZERS'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Union

from mmengine.runner import EpochBasedTrainLoop, IterBasedTrainLoop
from torch.utils.data import DataLoader

from mmrazor.registry import LOOPS


@LOOPS.register_module()
class DartsEpochBasedTrainLoop(EpochBasedTrainLoop):
    """EpochBasedTrainLoop for `Darts <https://arxiv.org/abs/1806.09055>`_

    In Darts, Two dataloaders are needed in the training stage. One
    (`dataloader`) is used to train the supernet and update its weights,
    another(`mutator_dataloader`) is only used to train and update the
    parameters of the supernet's architecture setting. In
    `DartsEpochBasedTrainLoop`, these dataloaders will be combined as a
    special dataloader, whose `data_batch` will contain both of the
    dataloaders' `data_batch`.

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or Dict):
            A dataloader object or a dict to build a dataloader for
            training the model.
        mutator_dataloader (Dataloader or Dict):
            A dataloader object or a dict to build a dataloader for
            training the parameters of model architecture.
        max_epochs (int): Total training epochs.
        val_begin (int): The epoch that begins validating.
            Defaults to 1.
        val_interval (int): Validation interval. Defaults to 1.
    """

    def __init__(self,
                 runner,
                 dataloader: Union[Dict, DataLoader],
                 mutator_dataloader: Union[Dict, DataLoader],
                 max_epochs: int,
                 val_begin: int = 1,
                 val_interval: int = 1) -> None:
        super().__init__(runner, dataloader, max_epochs, val_begin,
                         val_interval)
        if isinstance(mutator_dataloader, dict):
            self.mutator_dataloader = runner.build_dataloader(
                mutator_dataloader, seed=runner.seed)
        else:
            self.mutator_dataloader = mutator_dataloader
        self.multi_loaders = [self.dataloader, self.mutator_dataloader]

    def run_epoch(self) -> None:
        """Iterate one epoch."""
        self.runner.call_hook('before_train_epoch')
        self.runner.model.train()

        for idx, data_batch in enumerate(EpochMultiLoader(self.multi_loaders)):
            self.run_iter(idx, data_batch)

        self.runner.call_hook('after_train_epoch')
        self._epoch += 1


@LOOPS.register_module()
class DartsIterBasedTrainLoop(IterBasedTrainLoop):
    """IterBasedTrainLoop for `Darts <https://arxiv.org/abs/1806.09055>`_

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or Dict):
            A dataloader object or a dict to build a dataloader.
        mutator_dataloader (Dataloader or Dict):
            A dataloader object or a dict to build a dataloader for
            training the parameters of model architecture.
        max_iter (int): Total training iterations.
        val_begin (int): The iteration that begins validating.
            Defaults to 1.
        val_interval (int): Validation interval. Defaults to 1000.
    """

    def __init__(self,
                 runner,
                 dataloader: Union[Dict, DataLoader],
                 mutator_dataloader: Union[Dict, DataLoader],
                 max_iters: int,
                 val_begin: int = 1,
                 val_interval: int = 1000) -> None:
        super().__init__(runner, dataloader, max_iters, val_begin,
                         val_interval)
        if isinstance(mutator_dataloader, dict):
            self.mutator_dataloader = runner.build_dataloader(
                mutator_dataloader, seed=runner.seed)
        else:
            self.mutator_dataloader = mutator_dataloader
        multi_loaders = [self.dataloader, self.mutator_dataloader]
        self.multi_loaders = IterMultiLoader(multi_loaders)

    def run(self) -> None:
        """Launch training."""
        self.runner.call_hook('before_train')
        # In iteration-based training loop, we treat the whole training process
        # as a big epoch and execute the corresponding hook.
        self.runner.call_hook('before_train_epoch')
        while self._iter < self._max_iters:
            self.runner.model.train()

            data_batch = next(self.multi_loaders)  # type: ignore
            self.run_iter(data_batch)

            if (self.runner.val_loop is not None
                    and self._iter >= self.val_begin
                    and self._iter % self.val_interval == 0):
                self.runner.val_loop.run()

        self.runner.call_hook('after_train_epoch')
        self.runner.call_hook('after_train')


class EpochMultiLoader:
    """Multi loaders based on epoch."""

    def __init__(self, dataloaders: List[DataLoader]):
        self._dataloaders = dataloaders
        self.iter_loaders = [iter(loader) for loader in self._dataloaders]

    @property
    def num_loaders(self):
        """The number of dataloaders."""
        return len(self._dataloaders)

    def __iter__(self):
        """Return self when executing __iter__."""
        return self

    def __next__(self):
        """Get the next iter's data of multiple loaders."""
        data = tuple([next(loader) for loader in self.iter_loaders])

        return data

    def __len__(self):
        """Get the length of loader."""
        return min([len(loader) for loader in self._dataloaders])


class IterMultiLoader:
    """Multi loaders based on iter."""

    def __init__(self, dataloaders: Union[List[DataLoader], DataLoader]):
        self._dataloaders = dataloaders if isinstance(dataloaders,
                                                      list) else [dataloaders]
        self.iter_loaders = [iter(loader) for loader in self._dataloaders]
        self._epoch = 0

    @property
    def epoch(self):
        """The property of the class."""
        return self._epoch

    @property
    def num_loaders(self):
        """The number of dataloaders."""
        return len(self._dataloaders)

    def __next__(self):
        """Get the next iter's data of multiple loaders."""
        try:
            data = tuple([next(loader) for loader in self.iter_loaders])
        except StopIteration:
            self._epoch += 1
            for loader in self._dataloaders:
                if hasattr(loader.sampler, 'set_epoch'):
                    loader.sampler.set_epoch(self._epoch)
            self.iter_loader = [iter(loader) for loader in self._dataloaders]
            data = tuple([next(loader) for loader in self.iter_loaders])

        return data

    def __len__(self):
        """Get the length of loader."""
        return min([len(loader) for loader in self._dataloaders])



# Copyright (c) OpenMMLab. All rights reserved.
from .backward_tracer import BackwardTracer
from .loss_calculator import *  # noqa: F401,F403
from .parsers import *  # noqa: F401,F403
from .path import (Path, PathConcatNode, PathConvNode, PathDepthWiseConvNode,
                   PathLinearNode, PathList, PathNode, PathNormNode)

__all__ = [
    'BackwardTracer', 'PathConvNode', 'PathLinearNode', 'PathNormNode',
    'PathConcatNode', 'Path', 'PathList', 'PathNode', 'PathDepthWiseConvNode'
]



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABC, abstractmethod
from typing import Dict, Generic, Optional, TypeVar

from mmengine.model import BaseModule

CHOICE_TYPE = TypeVar('CHOICE_TYPE')
CHOSEN_TYPE = TypeVar('CHOSEN_TYPE')


class BaseMutable(BaseModule, ABC, Generic[CHOICE_TYPE, CHOSEN_TYPE]):
    """Base Class for mutables. Mutable means a searchable module widely used
    in Neural Architecture Search(NAS).

    It mainly consists of some optional operations, and achieving
    searchable function by handling choice with ``MUTATOR``.

    All subclass should implement the following APIs:

    - ``forward()``
    - ``fix_chosen()``
    - ``choices()``

    Args:
        module_kwargs (dict[str, dict], optional): Module initialization named
            arguments. Defaults to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(self,
                 alias: Optional[str] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg=init_cfg)

        self.alias = alias
        self._is_fixed = False
        self._current_choice: Optional[CHOICE_TYPE] = None

    @property
    def current_choice(self) -> Optional[CHOICE_TYPE]:
        """Current choice will affect :meth:`forward` and will be used in
        :func:`mmrazor.core.subnet.utils.export_fix_subnet` or mutator.
        """
        return self._current_choice

    @current_choice.setter
    def current_choice(self, choice: Optional[CHOICE_TYPE]) -> None:
        """Current choice setter will be executed in mutator."""
        self._current_choice = choice

    @property
    def is_fixed(self) -> bool:
        """bool: whether the mutable is fixed.

        Note:
            If a mutable is fixed, it is no longer a searchable module, just
                a normal fixed module.
            If a mutable is not fixed, it still is a searchable module.
        """
        return self._is_fixed

    @is_fixed.setter
    def is_fixed(self, is_fixed: bool) -> None:
        """Set the status of `is_fixed`."""
        assert isinstance(is_fixed, bool), \
            f'The type of `is_fixed` need to be bool type, ' \
            f'but got: {type(is_fixed)}'
        if self._is_fixed:
            raise AttributeError(
                'The mode of current MUTABLE is `fixed`. '
                'Please do not set `is_fixed` function repeatedly.')
        self._is_fixed = is_fixed

    @abstractmethod
    def fix_chosen(self, chosen: CHOSEN_TYPE) -> None:
        """Fix mutable with choice. This function would fix the choice of
        Mutable. The :attr:`is_fixed` will be set to True and only the selected
        operations can be retained. All subclasses must implement this method.

        Note:
            This operation is irreversible.
        """

    # TODO
    # type hint
    @abstractmethod
    def dump_chosen(self) -> CHOSEN_TYPE:
        ...

    @property
    @abstractmethod
    def num_choices(self) -> int:
        pass



# Copyright (c) OpenMMLab. All rights reserved.
from .mutable_channel import MutableChannel
from .one_shot_mutable_channel import OneShotMutableChannel
from .slimmable_mutable_channel import SlimmableMutableChannel

__all__ = [
    'OneShotMutableChannel', 'SlimmableMutableChannel', 'MutableChannel'
]



_base_ = ['./cwd_fpn_retina_r101_retina_r50_1x_coco.py']

teacher_ckpt = 'https://download.openmmlab.com/mmdetection/v2.0/gfl/gfl_r101_fpn_mstrain_2x_coco/gfl_r101_fpn_mstrain_2x_coco_20200629_200126-dd12f847.pth'  # noqa: E501
model = dict(
    architecture=dict(
        cfg_path='mmdet::gfl/gfl_r50_fpn_1x_coco.py', pretrained=False),
    teacher=dict(
        cfg_path='mmdet::gfl/gfl_r101_fpn_ms-2x_coco.py', pretrained=True),
    teacher_ckpt=teacher_ckpt)
